<?php
// commentinfo.php -- HotCRP helper class for comments
// HotCRP is Copyright (c) 2006-2016 Eddie Kohler and Regents of the UC
// Distributed under an MIT-like license; see LICENSE

class CommentInfo {
    public $commentId = 0;
    public $paperId;
    public $timeModified;
    public $timeNotified;
    public $timeDisplayed;
    public $comment;
    public $commentType = COMMENTTYPE_REVIEWER;
    public $replyTo;
    public $paperStorageId;
    public $ordinal;
    public $authorOrdinal;
    public $commentTags;
    public $commentRound;
    public $commentFormat;
    public $commentOverflow;

    static private $watching;
    static private $visibility_map = array(COMMENTTYPE_ADMINONLY => "admin", COMMENTTYPE_PCONLY => "pc", COMMENTTYPE_REVIEWER => "rev", COMMENTTYPE_AUTHOR => "au");


    function __construct($x = null, $prow = null) {
        $this->merge(is_object($x) ? $x : null, $prow);
    }

    private function merge($x, $prow) {
        global $Conf;
        if ($x)
            foreach ($x as $k => $v)
                $this->$k = $v;
        $this->commentId = (int) $this->commentId;
        $this->paperId = (int) $this->paperId;
        $this->commentType = (int) $this->commentType;
        $this->commentRound = (int) $this->commentRound;
        if ($Conf->sversion < 107 && $this->commentType >= COMMENTTYPE_AUTHOR)
            $this->authorOrdinal = $this->ordinal;
        $this->prow = $prow;
    }

    static public function fetch($result, $prow) {
        $cinfo = $result ? $result->fetch_object("CommentInfo", [null, $prow]) : null;
        if ($cinfo && !is_int($cinfo->commentId))
            $cinfo->merge(null, $prow);
        return $cinfo;
    }


    public static function echo_script($prow) {
        global $Conf, $Me;
        if (Ht::mark_stash("papercomment")) {
            $t = array("papercomment.commenttag_search_url=\"" . hoturl_raw("search", "q=cmt%3A%23\$") . "\"");
            if (!$prow->has_author($Me))
                $t[] = "papercomment.nonauthor=true";
            foreach ($Conf->resp_round_list() as $i => $rname) {
                $isuf = $i ? "_$i" : "";
                $wl = $Conf->setting("resp_words$isuf", 500);
                $j = array("words" => $wl);
                $ix = false;
                if ($Me->can_respond($prow, (object) array("commentType" => COMMENTTYPE_RESPONSE, "commentRound" => $i))) {
                    if ($i)
                        $ix = $Conf->message_html("resp_instrux_$i", array("wordlimit" => $wl));
                    if ($ix === false)
                        $ix = $Conf->message_html("resp_instrux", array("wordlimit" => $wl));
                    if ($ix !== false)
                        $j["instrux"] = $ix;
                }
                $t[] = "papercomment.set_resp_round(" . json_encode($rname) . "," . json_encode($j) . ")";
            }
            $Conf->echoScript(join($t, ";"));
        }
    }

    static private function commenttype_needs_ordinal($ctype) {
        return !($ctype & (COMMENTTYPE_RESPONSE | COMMENTTYPE_DRAFT))
            && ($ctype & COMMENTTYPE_VISIBILITY) != COMMENTTYPE_ADMINONLY;
    }

    private function ordinal_missing($ctype) {
        return self::commenttype_needs_ordinal($ctype)
            && !($ctype >= COMMENTTYPE_AUTHOR ? $this->authorOrdinal : $this->ordinal);
    }

    public function unparse_ordinal() {
        $is_author = $this->commentType >= COMMENTTYPE_AUTHOR;
        $o = $is_author ? $this->authorOrdinal : $this->ordinal;
        if (self::commenttype_needs_ordinal($this->commentType) && $o)
            return ($is_author ? "A" : "") . $o;
        else
            return null;
    }

    static public function unparse_html_id($cr) {
        global $Conf;
        $is_author = $cr->commentType >= COMMENTTYPE_AUTHOR;
        $o = $is_author ? $cr->authorOrdinal : $cr->ordinal;
        if (self::commenttype_needs_ordinal($cr->commentType) && $o)
            return ($is_author ? "cA" : "c") . $o;
        else if ($cr->commentType & COMMENTTYPE_RESPONSE)
            return $Conf->resp_round_text($cr->commentRound) . "response";
        else
            return "cx" . $cr->commentId;
    }

    private static function _user($x) {
        if (isset($x->reviewEmail))
            return (object) ["firstName" => get($x, "reviewFirstName"), "lastName" => get($x, "reviewLastName"), "email" => $x->reviewEmail];
        else
            return $x;
    }

    public function user() {
        return self::_user($this);
    }

    public function viewable_tags(Contact $user) {
        // NB caller must check can_view_comment_tags
        return Tagger::strip_nonviewable($this->commentTags, $user);
    }

    public function unparse_json($contact, $include_displayed_at = false) {
        global $Conf;
        if ($this->commentId && !$contact->can_view_comment($this->prow, $this, null))
            return false;

        // placeholder for new comment
        if (!$this->commentId) {
            if (!$contact->can_comment($this->prow, $this))
                return false;
            $cj = (object) array("pid" => $this->prow->paperId, "is_new" => true, "editable" => true);
            if ($this->commentType & COMMENTTYPE_RESPONSE)
                $cj->response = $Conf->resp_round_name($this->commentRound);
            return $cj;
        }

        // otherwise, viewable comment
        $cj = (object) array("pid" => $this->prow->paperId, "cid" => $this->commentId);
        $cj->ordinal = $this->unparse_ordinal();
        $cj->visibility = self::$visibility_map[$this->commentType & COMMENTTYPE_VISIBILITY];
        if ($this->commentType & COMMENTTYPE_BLIND)
            $cj->blind = true;
        if ($this->commentType & COMMENTTYPE_DRAFT)
            $cj->draft = true;
        if ($this->commentType & COMMENTTYPE_RESPONSE)
            $cj->response = $Conf->resp_round_name($this->commentRound);
        if ($contact->can_comment($this->prow, $this))
            $cj->editable = true;

        // tags
        if ($this->commentTags
            && $contact->can_view_comment_tags($this->prow, $this, null)) {
            if (($tags = $this->viewable_tags($contact)))
                $cj->tags = TagInfo::split($tags);
            if ($tags && ($cc = TagInfo::color_classes($tags)))
                $cj->color_classes = $cc;
        }

        // identity and time
        $idable = $contact->can_view_comment_identity($this->prow, $this, null);
        $idable_override = $idable || $contact->can_view_comment_identity($this->prow, $this, true);
        if ($idable || $idable_override) {
            $user = $this->user();
            $cj->author = Text::user_html($user);
            $cj->author_email = $user->email;
            if (!$idable)
                $cj->author_hidden = true;
        }
        if ($this->timeModified > 0 && $idable_override) {
            $cj->modified_at = (int) $this->timeModified;
            $cj->modified_at_text = $Conf->printableTime($cj->modified_at);
        } else if ($this->timeModified > 0) {
            $cj->modified_at = $Conf->obscure_time($this->timeModified);
            $cj->modified_at_text = $Conf->unparse_time_obscure($cj->modified_at);
            $cj->modified_at_obscured = true;
        }
        if ($include_displayed_at)
            // XXX exposes information, should hide before export
            $cj->displayed_at = (int) $this->timeDisplayed;

        // text
        if ($this->commentOverflow)
            $cj->text = $this->commentOverflow;
        else
            $cj->text = $this->comment;

        // format
        if (($fmt = $this->commentFormat) === null)
            $fmt = Conf::$gDefaultFormat;
        if ($fmt)
            $cj->format = (int) $fmt;
        return $cj;
    }

    public function unparse_text($contact, $no_title = false) {
        global $Conf;
        $x = "===========================================================================\n";
        if (!($this->commentType & COMMENTTYPE_RESPONSE))
            $n = "Comment";
        else if (($rname = $Conf->resp_round_text($this->commentRound)))
            $n = "$rname Response";
        else
            $n = "Response";
        if ($contact->can_view_comment_identity($this->prow, $this, false))
            $n .= " by " . Text::user_text($this->user());
        $x .= center_word_wrap($n);
        if (!$no_title)
            $x .= $this->prow->pretty_text_title();
        if ($this->commentTags
            && $contact->can_view_comment_tags($this->prow, $this, null)
            && ($tags = $this->viewable_tags($contact))) {
            $tagger = new Tagger($contact);
            $x .= center_word_wrap($tagger->unparse_hashed($tags));
        }
        $x .= "---------------------------------------------------------------------------\n";
        if ($this->commentOverflow)
            $x .= $this->commentOverflow;
        else
            $x .= $this->comment;
        return $x . "\n";
    }

    static public function unparse_flow_entry($crow, $contact, $trclass) {
        // See also ReviewForm::reviewFlowEntry
        global $Conf;
        $a = "<a href=\"" . hoturl("paper", "p=$crow->paperId#" . self::unparse_html_id($crow)) . "\"";
        $t = "<tr class='$trclass'><td class='pl_activityicon'>" . $a . ">"
            . Ht::img("comment24.png", "[Comment]", "dlimg")
            . '</a></td><td class="pl_activityid pnum">'
            . $a . ">#$crow->paperId</a></td><td class='pl_activitymain'><small>"
            . $a . " class=\"ptitle\">" . htmlspecialchars($crow->shortTitle);
        if (strlen($crow->shortTitle) != strlen($crow->title))
            $t .= "...";
        $t .= "</a>";
        $idable = $contact->can_view_comment_identity($crow, $crow, false);
        if ($idable || $contact->can_view_comment_time($crow, $crow))
            $time = $Conf->parseableTime($crow->timeModified, false);
        else
            $time = $Conf->unparse_time_obscure($Conf->obscure_time($crow->timeModified));
        $t .= ' <span class="barsep">·</span> ' . $time;
        if ($idable)
            $t .= ' <span class="barsep">·</span> <span class="hint">comment by</span> ' . Text::user_html(self::_user($crow));
        $t .= '</small><br /><a class="q" ' . substr($a, 3)
            . ">" . htmlspecialchars($crow->shortComment);
        if (strlen($crow->shortComment) < strlen($crow->comment))
            $t .= "...";
        return $t . "</a></td></tr>";
    }


    private function save_ordinal($cmtid, $ctype, $Table, $LinkTable, $LinkColumn) {
        global $Conf;
        $okey = "ordinal";
        if ($ctype >= COMMENTTYPE_AUTHOR && $Conf->sversion >= 107)
            $okey = "authorOrdinal";
        $q = "update $Table, (select coalesce(max($Table.$okey),0) maxOrdinal
    from $LinkTable
    left join $Table on ($Table.$LinkColumn=$LinkTable.$LinkColumn)
    where $LinkTable.$LinkColumn={$this->prow->$LinkColumn}
    group by $LinkTable.$LinkColumn) t
set $okey=(t.maxOrdinal+1) where commentId=$cmtid";
        Dbl::qe($q);
    }

    public function save($req, $contact) {
        global $Conf, $Now;
        if (is_array($req))
            $req = (object) $req;
        $Table = $this->prow->comment_table_name();
        $LinkTable = $this->prow->table_name();
        $LinkColumn = $this->prow->id_column();
        $req_visibility = get($req, "visibility");

        $is_response = !!($this->commentType & COMMENTTYPE_RESPONSE);
        if ($is_response && get($req, "submit"))
            $ctype = COMMENTTYPE_RESPONSE | COMMENTTYPE_AUTHOR;
        else if ($is_response)
            $ctype = COMMENTTYPE_RESPONSE | COMMENTTYPE_AUTHOR | COMMENTTYPE_DRAFT;
        else if ($req_visibility == "a" || $req_visibility == "au")
            $ctype = COMMENTTYPE_AUTHOR;
        else if ($req_visibility == "p" || $req_visibility == "pc")
            $ctype = COMMENTTYPE_PCONLY;
        else if ($req_visibility == "admin")
            $ctype = COMMENTTYPE_ADMINONLY;
        else if ($this->commentId && $req_visibility === null)
            $ctype = $this->commentType;
        else // $req->visibility == "r" || $req->visibility == "rev"
            $ctype = COMMENTTYPE_REVIEWER;
        if ($is_response ? $this->prow->blind : $Conf->is_review_blind(!!get($req, "blind")))
            $ctype |= COMMENTTYPE_BLIND;

        // tags
        if ($is_response) {
            $ctags = " response ";
            if (($rname = $Conf->resp_round_name($this->commentRound)) != "1")
                $ctags .= "{$rname}response ";
        } else if (get($req, "tags")
                   && preg_match_all(',\S+,', $req->tags, $m)) {
            $tagger = new Tagger($contact);
            $ctags = array();
            foreach ($m[0] as $text)
                if (($text = $tagger->check($text, Tagger::NOVALUE))
                    && !stri_ends_with($text, "response"))
                    $ctags[strtolower($text)] = $text;
            $tagger->sort($ctags);
            $ctags = count($ctags) ? " " . join(" ", $ctags) . " " : null;
        } else
            $ctags = null;

        // notifications
        $displayed = !($ctype & COMMENTTYPE_DRAFT);

        // query
        $text = get_s($req, "text");
        $q = "";
        $qv = array();
        if ($text === "" && $this->commentId) {
            $change = true;
            $q = "delete from $Table where commentId=$this->commentId";
        } else if ($text === "")
            /* do nothing */;
        else if (!$this->commentId) {
            $change = true;
            $qa = ["contactId, $LinkColumn, commentType, comment, commentOverflow, timeModified, replyTo"];
            $qb = [$contact->contactId, $this->prow->$LinkColumn, $ctype, "?", "?", $Now, 0];
            if (strlen($text) <= 32000)
                array_push($qv, $text, null);
            else
                array_push($qv, UnicodeHelper::utf8_prefix($text, 200), $text);
            if ($ctags !== null) {
                $qa[] = "commentTags";
                $qb[] = "?";
                $qv[] = $ctags;
            }
            if ($is_response) {
                $qa[] = "commentRound";
                $qb[] = $this->commentRound;
            }
            if ($displayed) {
                $qa[] = "timeDisplayed, timeNotified";
                $qb[] = "$Now, $Now";
            }
            $q = "insert into $Table (" . join(", ", $qa) . ") select " . join(", ", $qb) . "\n";
            if ($is_response) {
                // make sure there is exactly one response
                $q .= " from (select $LinkTable.$LinkColumn, coalesce(commentId, 0) commentId
                from $LinkTable
                left join $Table on ($Table.$LinkColumn=$LinkTable.$LinkColumn and (commentType&" . COMMENTTYPE_RESPONSE . ")!=0 and commentRound=$this->commentRound)
                where $LinkTable.$LinkColumn={$this->prow->$LinkColumn} limit 1) t
        where t.commentId=0";
            }
        } else {
            $change = ($this->commentType >= COMMENTTYPE_AUTHOR) != ($ctype >= COMMENTTYPE_AUTHOR);
            if ($this->timeModified >= $Now)
                $Now = $this->timeModified + 1;
            // do not notify on updates within 3 hours
            $qa = "";
            if ($this->timeNotified + 10800 < $Now
                || (($ctype & COMMENTTYPE_RESPONSE)
                    && !($ctype & COMMENTTYPE_DRAFT)
                    && ($this->commentType & COMMENTTYPE_DRAFT)))
                $qa .= ", timeNotified=$Now";
            // reset timeDisplayed if you change the comment type
            if ((!$this->timeDisplayed || $this->ordinal_missing($ctype))
                && $text !== "" && $displayed)
                $qa .= ", timeDisplayed=$Now";
            $q = "update $Table set timeModified=$Now$qa, commentType=$ctype, comment=?, commentOverflow=?, commentTags=? where commentId=$this->commentId";
            if (strlen($text) <= 32000)
                array_push($qv, $text, null);
            else
                array_push($qv, UnicodeHelper::utf8_prefix($text, 200), $text);
            $qv[] = $ctags;
        }

        $result = Dbl::qe_apply($q, $qv);
        if (!$result)
            return false;
        $cmtid = $this->commentId ? : $result->insert_id;
        if (!$cmtid)
            return false;

        // log
        $contact->log_activity("Comment $cmtid " . ($text !== "" ? "saved" : "deleted"), $this->prow->$LinkColumn);

        // ordinal
        if ($text !== "" && $this->ordinal_missing($ctype))
            $this->save_ordinal($cmtid, $ctype, $Table, $LinkTable, $LinkColumn);

        // reload
        if ($text !== "") {
            $comments = $this->prow->fetch_comments("commentId=$cmtid");
            $this->merge($comments[$cmtid], $this->prow);
            if ($this->timeNotified == $this->timeModified) {
                self::$watching = $this;
                $this->prow->notify(WATCHTYPE_COMMENT, "CommentInfo::watch_callback", $contact);
                self::$watching = null;
            }
        } else {
            $this->commentId = 0;
            $this->comment = "";
            $this->commentTags = null;
        }

        return true;
    }

    static function watch_callback($prow, $minic) {
        $ctype = self::$watching->commentType;
        if (($ctype & COMMENTTYPE_RESPONSE) && ($ctype & COMMENTTYPE_DRAFT))
            $tmpl = "@responsedraftnotify";
        else if ($ctype & COMMENTTYPE_RESPONSE)
            $tmpl = "@responsenotify";
        else
            $tmpl = "@commentnotify";
        if ($minic->can_view_comment($prow, self::$watching, false)
            // Don't send notifications about draft responses to the chair,
            // even though the chair can see draft responses.
            && ($tmpl !== "@responsedraftnotify" || $minic->act_author_view($prow)))
            HotCRPMailer::send_to($minic, $tmpl, $prow, array("comment_row" => self::$watching));
    }
}
